# Cyc.pm: module for interfacing with Cyc server 
#
# Copyright (c) 2000 - 2001 Cycorp, Inc.  All rights reserved.
#
# Developed by Charles Sutton for QuA and revised by Tom O'Hara
#
# Notes:
#
# The code was written assuming a newline terminates the data being sent. 
# Thus multiple-line results are not supported. However, sometimes the
# results returned from Cyc include embedded newlines, in particular
# sometimes in term comments, and also for some defstructure print methods.
# To work-around this problem, there is a customization of default-api-output-protocol
# in api-kernel.lisp to convert newlines (\n) to carriage returns (\r).
# This work-around is loaded in by the perl scripts. See NL_support.perl
# and NL_support.lisp in
#   ~/cvs/head/cycorp/cyc/top/scripts/active-indexing

package Cyc;

BEGIN {
    require 5.004;
    use Carp;
    use Socket;
    use FileHandle;
    use vars     qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
    $VERSION     = "0.8";
}


use vars qw($debug);
$debug = 0;


#################################################################
#
#  General functions for connecting to FI port, interpreting
#  responses, etc.
#
#################################################################

sub new {
    print STDERR "Cyc::new(@_)\n" if ($debug);
    my $class = shift;
    my %args = @_;

    my $host = $args{'host'};
    my $port = $args{'port'};
    $host ||= 'localhost';
    $port ||= '3601';
	
    return bless { 'fh' => &open_cyc_socket ($host, $port) }, $class;
}


######################################################
#   DESTROY
#      Class finalizer.
#      Needed to close the cyc socket, and also to prevent
#      Perl from trying to autoload DESTROY.  And Cyc::AUTOLOAD
#      automatically sends an undefined function to Cyc as a
#      SubL function call.  This would be bad.
#
sub DESTROY {
    print STDERR "Cyc::DESTROY(@_)\n" if ($debug);
    my $self = shift;

    print { $self->{'fh'} } "(fi-quit)";
    close $self->{'fh'};
}


use vars qw($socket_fh);
$socket_fh = 'cyc00';  # Maintains what we should call the next filehandle

#######################################################
#  run_cyc
#    Runs a cyc C executable in a child process.
#    Returns when the executable has loaded.
#   Inputs:
#      executable : name of executable
#      arguments : string of command line arguments
#      run_dir : directory to chdir to (optional)
#   Return value:
#      PID of process
#      Returns undef on error.
#
#   This is not a class method.
#
sub run_cyc {
    my ($executable, $arguments, $run_dir, $rsh) = @_;

    $socket_fh++;

    my $cmdline;
    if ($run_dir) {
	$cmdline = "cd $run_dir; ";
    }
    $cmdline .= "$executable $arguments 2>&1 -b";   

    if ($rsh) {
	$cmdline = "rsh $rsh \"$cmdline\" |";
    } else {
	$cmdline = "$cmdline |";
    }
    if ($debug) { print STDERR "Starting $cmdline\n"; }
    my $pid = open $socket_fh, $cmdline;
    return undef unless $pid;

    if ($debug) { print STDERR "PID is: $pid\n"; }

    while (<$socket_fh>) {
	if ($debug) { print STDERR "$_"; }
	if (/^Done./) { last; }
	if (/^\? /) {
	    close $socket_fh;
	    $pid = $undef;
	    last;
	}
    }

    return $pid;
}


#######################################################
#  open_cyc_socket
#    Opens a TCP connection to Cyc.  This is not a method
#    of the Cyc object.
#
#   Inputs: hostname
#           port
#   Output: the filehandle which the connection was opened on.
#
sub open_cyc_socket {
    my ($remote, $port) = @_;
    my ($iaddr, $paddr, $proto, $line);

    $iaddr = inet_aton($remote);
    $paddr = sockaddr_in($port, $iaddr);
    $proto = getprotobyname('tcp');

    $socket_fh++;

    socket($socket_fh, PF_INET, SOCK_STREAM, $proto) 
	or croak "socket: $!";
    connect($socket_fh, $paddr) 
	or croak "connect: $!";

    $socket_fh->autoflush (1);

    $socket_fh;
} 


######################################################
#  fi_send
#    Sends a query through an already-opened socket to a
#    Cyc FI server.
#   Inputs: cyc object that does the request,
#           string containing fi command
#   Output:   Returns a string containing a response.
#             Use Cyc::fi_error to determine whether the response was successful.
#
sub fi_send {
    my ($object, $command) = @_;
    my $response;
    my $fh = $object->{'fh'};
    my $result;

    # Make sure in line input mode
    my $input_record_separator = $/;
    $/ = "\n";

    print STDERR "SENDING $command\n" if ($debug);
    print $fh "$command\n";
    $response = $object->read_response();
    if ($response =~ /^200\s+([^\000]*)/) {
	$response = $1;
	$response =~ s/^\s+//;  # Remove leading whitespace
	$response =~ s/\s+$//;  # Remove trailing whitespace, including \n & \r
	print STDERR "RESULT  $response\n" if ($debug);
	$result = $response;
    } else {
	print STDERR "\nERROR: $response" if ($debug);
	$result = "ERROR $response";
    }

    # Restore previous line input mode
    $/ = $input_record_separator;

    return $result;
} 


# has_open_quote(string): determines whether the string has an open quote
# (excluding escape quotes)
#
sub has_open_quote {
    my($string) = @_;
    my($count) = 0;
    my($escaped) = 0;

    # Loop through the string, counting number of unescaped quotes
    for (my $i = 0; $i < length($string); $i++) {
	my($char) = substr($string, $i, 1);
	if (($char eq "\"")  && (! $escaped)) {
	    $count++;
	}
	elsif ($char eq "\\") {
	    $escaped = 1;
	}
	else {
	    $escaped = 0;
	}
    }

    # Determine open-quotedness by whether an odd number of quotes found
    my($is_quoted) = &is_odd($count);

    return ($is_quoted);
}


#########################################################
# Cyc::read_response(): Reads in the next block of text from the TCP socket.
#
# This is usually a line, but it could be several in case of embedded newlines.
# NOTE: This assumes that the result is output from Lisp interpreter and thus
# only quoted strings will be split across lines.
# TODO: verify that this is in accordance with FTP protocol
#
sub read_response {
    my($object) = @_;
    my($fh) = $object->{'fh'};
    my($ok) = 1;

    # Read in the next line of data
    my $response = <$fh>;
    $response = "<unexpected end of file>" if (!defined($response));
    print STDERR "response: $response" if ($debug > 1);

    # Read in additional lines while inside quoted string
    while (&has_open_quote($response) && $ok) {
	print STDERR "waiting for subsequent reply\n" if ($debug > 1);
	my($new_response);
	if ($object->is_socket_blocked()) {
	    print STDERR "unexpected TCP error (blocked socket)\n";
	    $ok = 0;
	}
	else {
	    $new_response = <$fh>;
	}
	if (!defined($new_response)) {
	    $new_response = "<unexpected end of file>";
	    $ok = 0;
	}
	print STDERR "new_response: $new_response" if ($debug > 1);
	$response .= $new_response;
    }

    return ($response);
}

# is_blocked: determines whether the Cyc socket is blocked
#
use Fcntl;
#
sub is_socket_blocked {
    my($object) = @_;
    my($fh) = $object->{'fh'};
    my($blocked) = 0;
    my($flags);

    # fcntl() performs a variety of functions on open descriptors:
    #
    # F_GETFL        Get descriptor status flags (see fcntl(5) for
    #                definitions).
    #
    $flags = fcntl($fh, &F_GETFL, 0);
    print STDERR "fcntl($fh, &F_GETFL, 0) => $flags\n" if ($debug > 1);
    if ($flags eq "2") {
	$blocked = 1;
    }

    return ($blocked);
}


##########################################################
#  fi_error
#    Returns T iff the given response from Cyc::fi_send indicates
#    an error.  May be called as a class method or an instance method.
#
sub fi_error {
    my ($class, $response) = @_;

    substr ($response, 0, 5) eq "ERROR";
}


#####################################################
#  lisp_read
#    A recursive-descent parser that reads Lisp expressions.
#    At present, this only handles conses, symbols. and strings.
#    Lists are converted into Perl arrays.  Nested lists are convered
#       into nested arrays.  (Should work for arbitrarily nested lists.)
#       Dotted conses are treated exactly like undotted conses.
#       NIL is treated as an empty array.
#    Atoms are returned by a string containing their printed
#       representation.
#
#    Input: A string containing a Lisp sexp
#    Returns: In list context, a list of (rest of line), (parsed-expression)
#             Parsed expression is a string (for a non-nil atom)
#                                  or an array ref (for a list--including NIL)
#             In scalar context, returns only the parsed expression           
#
sub lisp_read {
    my $class = shift;
    my $line = shift;
    my $sem;
    print STDERR "lisp_read($line)\n" if ($debug > 1);

    $line =~ s/\s*//;
 
    if (substr ($line, 0, 1) eq ".") {
	$line = substr ($line, 1, length ($line) - 1);
	$line =~ s/\s*//;
    }

    if ($line =~ /^\(/) {
	$line = substr ($line, 1, length ($line) - 1);
	($line, $sem) = &lisp_read_list ($class, $line);
    } elsif ($line =~ /^NIL/) {
	$line = substr ($line, 3, length ($line) - 3);
	$sem = [];
    } elsif ($line =~ s/^\s*(\"((\\\"|[^\"])*)\")//) {
	$sem = $1;
    } else {
	$line =~ s/([^\)\s]+)//;
	$sem = $1;
    }
    print STDERR "lisp_read => $sem\n" if ($debug > 2);

    if (wantarray()) {
	return ($line, $sem);   # List context
    } elsif (defined wantarray()) {
	return $sem;            # Scalar context
    } else { 
	return;                 # Void context
    }
}


#######################################################
#  lisp_read_list
#
#    Reads a Lisp list.  This is used by lisp_read, which should
#    be used for more general reading.
#
#    Input: The text of the list (not counting opening paren)
#    Returns: A list of (rest of line), (parsed-expression)
#             Parsed expression will be an arrayref where each element
#              is a string (for a symbol) or an array (for a list).
#
sub lisp_read_list {
    my $class = shift;
    my $line = shift;
    my (@sem, $sem_temp);

    while ($line) {
	last if ($line =~ s/^\)//);
	($line, $sem_temp) = &lisp_read ($class, $line);
	push @sem, $sem_temp;
    }

    ($line, \@sem);
}


######################################################
#  uncyclify
#   If a given string begins in #$, remove it
#
sub uncyclify {
    my ($class, $temp) = @_;

    $temp =~ s/^#\$//;
    $temp;
}


#########################################################
#  cyclify_string
#    If a given string does not begin in #$, add it.
sub cyclify_string {
   my $temp = shift;

  ($temp =~ /^\#\$/) ? $temp : '#$' . $temp;
}

###########################################################
#
#  Perl wrappers to FI functions.
#
###########################################################


######################################################
#  AUTOLOAD
#    Treats any undefined functions as FI calls.
#
sub AUTOLOAD {

    my $class = shift;
    my @stack = reverse @_;

    carp "Warning: Treating function $AUTOLOAD as a Cyc FI call" if ($debug);
    my $function = $AUTOLOAD;
    $function =~ s/^.*:://;
    $function =~ s/_/\-/g;
    $function = "($function";

    while (@stack) {
	my $elem = pop @stack;
	if (!ref ($elem)) {
	    $function = "$function $elem ";
	} elsif (ref ($elem) eq 'ARRAY') {
	    $function .= '(';
	    push @stack, ')';
	    push @stack, reverse (@$elem);
	} elsif (ref ($elem eq 'HASH')) {
	    push @stack, reverse (values %$elem);
	}
    }
    
    $function .= ')';

#    print $function;
    $response = &fi_send ($class, $function);

}

######################################################
#  find_constant
#    Returns constant exists in the Cyc KB, returns
#    its name, otherwise returns false.
#
sub find_constant {
    my $self = shift;
    my $constant = shift;
    my $response = $self->fi_send ("(find-constant \"$constant\")");

    ($response ne 'NIL') ? $response : 0; 
}

######################################################
#  valid_constant_name
#    Returns true iff the input is a valid constant name.
#    TODO: Should be implemented in Perl
#
sub valid_constant_name {
    my $self = shift;
    my $constant = shift;
    my $response = $self->fi_send ("(valid-constant-name \"$constant\")");

    ($response ne 'NIL') ? $response : 0; 
}

######################################################
#  constant_name_case_collisions
#    Input: $constant_name = string
#    Output: If $constant_name collides with an existing constant
#            case-insenstively, returns a reference to an array of
#             colliding constants.
#            Otherwise, returns false.
#
sub constant_name_case_collisions {
    my $self = shift;
    my $constant = shift;
    my $response = $self->fi_send ("(constant-name-case-collisions \"$constant\")");
    
    ($response ne 'NIL') ? (&lisp_read ($response))[1] : 0;
	    
}


#####################################################
#  fi_create
#    Creates a constant in the Cyc KB.  Returns
#    the result of the FI call, which can be checked
#    with fi_error.
#
sub fi_create {
    my $self = shift;
    my $constant_name = shift;
    my $response = $self->fi_send ("(fi-create \"$constant_name\")");

    $response;
}



#########################################################################
#
#  Below contains Query Augmentation functionality.
#
#########################################################################


#########################################################################
#  qa_query
#    Does a Query Augmentation QUERY command from Perl.  Assumes that
#    the socket corresponding to this
#   Inputs: the string to query, and the session id (defaults to 1)
#   Outputs: an array reference, where each element is of the form
#              [ label-type, label-id, printstring ]
#            If QuA does not know the query, returns the string 'NO_HELP'
#            If QuA returns an error, returns the string 'ERROR: XXXXXXX'
#             where XXXXXXXX is the error message.
#  
sub qa_query {
    my ($object, $query, $session_id) = @_;
    my (@label_list, $response);
    my $fh = $object->{'fh'};

    $session_id ||= 1;

    print $fh "query \"$query\" $session_id\n";
    $response = <$fh>;
#    print "$response\n";
    if (substr ($response, 0, 3) eq "224") {
	$response = <$fh>;  # Catch the trailing .
	return 'NO_HELP';
    } elsif ((substr ($response, 0, 1) eq "4") ||
             (substr ($response, 0, 1) eq "5") ||
	     (substr ($response, 0, 3) eq "201")) {
	return "ERROR: $response";
    } else {
	until (substr($response, 0, 1) eq ".") {
	    $response = <$fh>;
#    print "$response\n";

	    $response =~ /^(\w\w\w) (\d+) \"([^\"]+)\"\s*/; 
	    if (($1 eq "DIS") || ($1 eq "SUB") || ($1 eq "CON")) {
		push @label_list, [ $1, $2, $3 ];
	    }
	}
    }

    \@label_list;
}

#########################################################################
#  qa_register
#
#    Does a Query Augmentation REGISTER command from Perl.
#    Assumes that the connection by Cyc which belongs to the calling
#     object is using the QA protocol.
#
#   Inputs: the session-id, the label-type, and the label-id
#   Outputs: an array, where each element is of the form
#              [ label-type, label-id, printstring ]
#            The first such entry is the registered node.
#      On error, returns a string 'ERROR: XXXXX', where XXXXX is the error message
#       returned by the QuA server.
#  
sub qa_register {
    my ($object, $sess_id, $label_type, $label_id) = @_;
    my (@label_list, $response);
    my $fh = $object->{'fh'};

    print $fh "register $sess_id $label_type $label_id\n";
    $response = " ";

    if ((substr ($response, 0, 1) eq "4") ||
	(substr ($response, 0, 1) eq "5") ||
	(substr ($response, 0, 3) eq "201")) {
	return 'ERROR: ' . substr ($response, 4);
    } else {
	
	until ((substr($response, 0, 1) eq ".") || 
	       (substr($response, 0, 1) eq "5") ||
	       (substr($response, 0, 3) eq "201") ||
	       (substr($response, 0, 3) eq "440")) {
	    $response = <$fh>;

	    if (substr($response, 0, 3) eq "REG") {
		$response =~ /^REG (\w\w\w) (\d+) \"([^\"]+)\"/; 
	    } else {
		$response =~ /^(\w\w\w) (\d+) \"([^\"]+)\"/; 
	    }

	    if (($1 eq "DIS") || ($1 eq "SUB") || ($1 eq "CON")) {
		push @label_list, [ $1, $2, $3 ];
	    }
	}
    }

    \@label_list;
}


# id_odd(number): indicates whether the number is odd or not
#
sub is_odd {

    my $num = shift;

    use integer;
    my $int_div = $num / 2;
    no integer;
    my $float_div = $num / 2;

    $int_div != $float_div;

}

1;
__END__

=head1 NAME

Cyc - Perl extension for connection to a Cyc server

=head1 SYNOPSIS

  use Cyc;
  $cyc = Cyc->new ( host => 'hostname',
		    port => '3601' );

  #
  # Both of these send to Cyc (fi-ask '(#$isa #$Charles #$HumanCyclist) #$CyclistsMt)
  #
  $response = $cyc->fi_send ('(fi-ask \'(#$isa #$Charles #$HumanCyclist) #$CyclistsMt)');
  $response2 = $cyc->fi_ask ('\'(#$isa #$Charles #$HumanCyclist) #$CyclistsMt');

  #
  # Sends (fi-ask '(#$genls ?X #$ComputerProgram) #$ComputerGMt)
  #
  $response3 = $cyc->fi_ask ('\'(#$genls ?X #$ComputerProgram) #$ComputerGMt');

  #
  # Sends (mapcar #'not '(a b NIL c))
  #
  $response4 = $cyc->mapcar ("#'not", "'", [ 'a', 'b', 'NIL', 'c' ]);

  #
  # Parsing the ask results
  #
  foreach $binding_set ($cyc->lisp_read ($response3)) {
      print "More bindings: \n";
      foreach $binding_pair ($binding_set) {
	  # This won't handle NATs correctly.  To do that, you'll have to
	  # look at the entire array from $binding_pair->[1] on, handling
	  # nested arrays.  This would be a good candidate for a new
	  # library function. :-)
	  print "BINDING of $binding_pair->[0] to $binding_pair->[1]\n";
      }
      print "\n";
  }

=head1 DESCRIPTION

This module contains methods to connect to a Cyc server, perform operations,
and parse results that are in the form of Lisp lists.

Any undefined function (e.g. Cyc::fi_ask and Cyc::mapcar, above) will be
forwarded to the Cyc server as a Lisp function call, with the substitution
s/_/\-/g run on the name of the undefined Perl function.  Arguments to the Perl
function will be passed along as expected to the Lisp function call.
If some of the arguments are array references, these will be converted into
nested lists.

Here are the main external functions:

=over 4

=item $cyc->new ( host => $hostname, port => $port );

Creates a new Cyc object, connect to a Cyc server running on $hostname at port
$port.  $hostname defaults to 'localhost' and $port defaults to 3601.

=item $cyc->fi_send ($command)

Sends the string $command to the Cyc server.  This function may block.
Returns a string contain the response from the Cyc server.  The numeric response
code is removed.   

=item $cyc->fi_error ($response)

Returns true if $response, which was returned by $cyc->fi_send, indicates an
error.

=item $cyc->lisp_read ($string)

Parser to read the string $string as a Lisp sexp.  At present, this
only handles conses, symbols. and strings.  Lists are converted into
Perl arrays.  Nested lists are convered  into nested arrays.  (Should
work for arbitrarily nested lists.)   Dotted conses are treated
exactly like undotted conses.  NIL is treated as a reference to an empty array.  
Atoms are returned by a string containing their printed 
representation.  In list context, returns a list of (rest of line), (parsed-expression).
In scalar context, simply returns the parser output: a string (for an atom) or array
reference (for a list).

=item $cyc->uncyclify ($string)

If $string begins in #$, remove it.  Returns the uncyclified string.

=back

The following functions provide convenient wrappers to FI calls.

=over 4

=item  $cyc->find_constant ($constant_name)

If a constant named $constant_name exists in the Cyc KB, returns 
its name, otherwise returns false.

=item $cyc->valid_constant_name ($name)

Returns true iff the $name would be a valid name for a Cyc constant.

=item $cyc->constant_name_case_collisions ($constant_name) 

If $constant_name collides with an existing constant 
case-insenstively, returns a reference to an array of colliding
constants.  Otherwise, returns false.  

=item $cyc->fi_create ($constant_name)

Creates a constant in the Cyc KB name $constant_name.  Returns 
the result of the FI call, which can be checked for an error
condition with $cyc->fi_error.

=back

The following functions provide access to the Query Augmentation
functionality.

=over 4

=item $cyc->qa_query ($query, $session_id)

Does a Query Augmentation QUERY command from Perl. 
Uses the string given in $query, and the session id $session_id.
If the query command is sucessful, returns an array reference,
where each element is of the form

    [ label-type, label-id, printstring ]

If QuA does not understand the query, returns the string 'NO_HELP'.
If QuA returns an error, returns the string 'ERROR: XXXXXXX'
where XXXXXXXX is the error message.

Assumes that the connection used by $cyc is using the QA protocol.
To start using the QA protocol, use the command $cyc->qa_protocol();

=item $cyc->qa_register ($session_id, $label_type, $label_id)

Does a Query Augmentation REGISTER command from Perl.
Registers the label of type $label_type and id $label_id in
session $session_id.
If the command is successful, returns an array ref, where each element is of the form

    [ label-type, label-id, printstring ]

The first such entry is the registered node.
On error, returns a string 'ERROR: XXXXX', where XXXXX is the error message
returned by the QuA server.

Assumes that the connection used by $cyc is using the QA protocol.
To start using the QA protocol, use the command $cyc->qa_protocol();

=over 4

=item Cyc::run_cyc ($executable, $arguments, [$run_dir], [$rsh_machine]),

Starts the Cyc executable $executable with command line arguments
$arguments.  Will block until the executable is fully-loaded.
If $run_dir is specified, will chdir to $run_dir first.  If $rsh_machine
is specified, will use rsh to run command on remote machine (this is 
unlikely to work on non-POSIX systems).  Returns PID of child process,
or undef on error.

Unfortunately, you'll probably have to kill the iamge using
kill 'KILL', $cyc_pid, or else telnet in and do a (quit).


=back

=head1 AUTHOR

Charles Sutton, sutton@cyc.com

=head1 SEE ALSO

perlreftut(1), Cyc FI documentation.

=cut

