# *-*-perl-*-*
eval 'exec perl -Ssw $0 "$@"'
    if 0;
#!/usr/bin/perl -sw
#
# count_it.perl: script to count the occurrences of a pattern in the input
#
# NOTE: This is a simple script that turns out to be very useful
# for a variety of tasks, especially in corpus analysis.
#
# examples:  
#
# tabulating most commonly used commands:
#
#   $ history | count_it.perl "^\s*\d+\s*(\S+)" - | less
#   ps_mine 13
#   w       11
#   gr      7
#   top     6
#   ...
# 
# tabulating part-of-speech usage for particular words
#
#   $ count_it "(outside\/\S+)" ~/OpenMind/data/omcsraw.tag 
#   outside/IN	502
#   outside/RB	137
#   outside/NN	53
#   outside/JJ	19
#
# Copyright (c) 2000-2001 Cycorp, Inc.  All rights reserved.
# Portions Copyright (c) 1996-1999 Tom O'Hara (at NMSU).  All rights reserved.
#

# Load in the common module, making sure the script dir is in Perl's lib path
BEGIN { my $dir = `dirname $0`; chomp $dir; unshift(@INC, $dir); }
require 'common.perl';
require 'extra.perl';

# Check the command-line options
# Each variable initialized corresponds to -var=value commandline option
&init_var(*i, &FALSE);			# ignore case in the pattern matching
&init_var(*ignore_case, $i);		# alias for -i option
&init_var(*foldcase, $i);		# fold (convert) text to lowercase
&init_var(*fold, $i);			# alias for -foldcase
&init_var(*preserve, ! ($foldcase || $fold));	# preserve the case of text matching pattern
&init_var(*para, &FALSE);		# apply the pattern to paragraphs not lines
&init_var(*one_per_line, &FALSE);	# only count one instance of the pattern per line
&init_var(*freq_first, &FALSE);		# put frequency counts first (ie, <freq><tab><data>)
&init_var(*alpha, &FALSE);		# alphabetical sort
&init_var(*compact, &FALSE);		# compact all whitespace sequences
&init_var(*cumulative, &FALSE);		# include column for cumulative counts
&init_var(*occurrences, &FALSE);	# the tags being counted are actually occurrence counts
&init_var(*occurrence_field, $occurrences);	# field giving occurence count
&init_var(*percents, &FALSE);		# shows the relative percents

# Get the pattern and options from the command line
# Process the command-line options
if (!defined($ARGV[0])) {
    my($options) = "options = [-i(gnore)] [-alpha] [-freq_first] [-para] [-preserve] [-foldcase] [-compact]\n";
    my($example) = "examples:\n\nls | $script_name '\\.([^\\.]+)\$'\n\n";
    $example .= "$0 '(outside\/\\S+)' omcsraw.tag\n\n";
    my($note) = "";
    $note .= "notes:\n\nThe patterns are regular expressions using Perl's extensions\n";
    $note .= "See manual page for details (e.g., 'man perlre')\n\n";
    $note .= "To tabulate only parts of the pattern, use parenthesis as illustrated above.\n";

    die "\nusage: $script_name [options] pattern file ...\n\n$options\n\n$example\n$note\n";
}
$pattern = $ARGV[0];
shift @ARGV;

# TODO: handle multiple patterns per line (eg, set line break to null)
# NOTE: checks against pattern need to occur prior to modification (eg, paren addition)
if ($pattern =~ /^\^/) {
    $one_per_line = &TRUE;
}

# Put grouping parenthesis around pattern, if none present
if ($pattern !~ /\(.*\)/) {
    $pattern = "(" . $pattern . ")";
}

# Process each line of the input stream
&debug_out(&TL_VERBOSE, "searching for '%s'; one_per_line=%d; ignore_case=%d...\n", 
	   $pattern, $one_per_line, $ignore_case);
$num = 0;
if ($para) {
    $/ = "";			# paragraph input mode
}
my($total_count) = 0;
while (<>) {
    # NOTE: a chop isn't performed to allow for using the newline in a pattern.
    # For instance, this is sometimes more convenient than $ when using csh.
    &dump_line();

    # Perform optional transformations
    # - In paragraph mode treat newlines as spaces unless mentioned in the pattern.
    # - All whitespace sequences to single spaces
    if ($para && ($pattern !~ /\n/)) {
	s/\n/ /g;
    }
    if ($compact) {
	s/\s+/ /g;
    }

    $text = $_;
    while (($text ne "") && ($text ne "\n")) {
	&debug_out(&TL_VERY_VERBOSE, "text=%s\n", $text);
	my($found) = &FALSE;

	# Try to extract tag from the text
	if (($ignore_case == 1) && ($text =~ /$pattern/i)) {
	    $tag = $1;
	    $tag = &to_lower($tag) unless ($preserve);
	    $text = $';
	    $found = &TRUE;
	}
	elsif (($ignore_case == 0) && ($text =~ /$pattern/)) {
	    $tag = $1;
	    $text = $';
	    $found = &TRUE;
	}

	# Update the hash table of the patterns
	if ($found) {
	    &debug_out(&TL_VERY_VERBOSE, "adding: %s\n", $tag);
	    $num = $num + 1;
	    my($count) = 1;
	    if ($occurrence_field) {
		$count = eval "$${occurrence_field}";
	    }
	    &incr_entry(*count, $tag, $count);
	    $total_count += $count;
	}

	# Stop when not found or if just one match per line
	if ((! $found) || ($one_per_line)) {
	    $text = "";
	}
    }
}
&debug_out(&TL_VERBOSE, "$num patterns found\n");
&debug_out(&TL_VERBOSE, "total occurrence count is ${total_count}\n") if ($occurrences);

my($sort_function) = ($alpha ? \&sorted_hash_keys_alphabetic : \&sorted_hash_keys_reverse_numeric);
if ($occurrences) {
    $sort_function = \&sorted_hash_keys_numeric_keys;
}
my($cumulative_tag_count) = 0;
foreach $tag (&$sort_function(\%count)) {
    my($tag_count) = &get_entry(*count, $tag, 0);
    if ($tag_count != 0) {
	$tag =~ s/\n$//;
	if ($freq_first) {
	    print $tag_count, "\t", $tag;
	}
	else {
	    print $tag, "\t", $tag_count;
	}
	if ($percents) {
	    print "\t", &round($tag_count / $total_count);
	}

	# Show optional cumulative count column
	if ($cumulative) {
	    my($occurrence_count) = ($occurrences ? $tag : 1);
	    $cumulative_tag_count += ($tag_count * $occurrence_count);
	    print "\t${cumulative_tag_count}";
	    if ($percents) {
		print "\t", &round($cumulative_tag_count / $total_count);
	    }
	}
	print "\n";
    }
}

print "\n";
