# cyc_helper.perl: helper routines for using the Cyc.pm module
#
# usage:
#    require 'cyc_helper.perl';
#    ...
#    &init_cyc();
#
# Copyright (c) 2000 - 2001 Cycorp, Inc.  All rights reserved.
#

local($cyc);

&init_var(*host, "localhost");		# host name of the Cyc server
&init_var(*port, 3601);			# use 3621 for localhost20, 3641 for localhost40, etc.
&init_var(*debug, &FALSE);		# debug flag for Cyc module
&init_var(*uncyclify_results, &FALSE);	# allows automatic uncyclifying (eg, removing #$'s)
&init_var(*profile, &FALSE);		# profile the Cyc calls
&init_var(*USER, "guest");		# Unix user ID
&init_var(*user, $USER);		# alias for above
&init_var(*cyclist, "");		# cyclist to use

# Hsh tables for use in profiling
my(%function_calls, %function_times);
my($gettimeofday) = &find_library_file("gettimeofday");

sub CYC_OPTIONS {"[-host=hostname] [-port=num]";}
sub CYC_OPTIONS_NOTES  {"For localhost20, use port 20 (or 3621), localhost40, 40 (or 3641), etc.";}

# init_cyc([port]): Initialize Cyc TCP connection
#
sub init_cyc {
    &debug_out(&TL_VERBOSE, "init_cyc(@_): port=$port\n");
    my($server_port) = @_;
    $server_port = $port if (!defined($server_port));

    my($OK) = &FALSE;
    if ($server_port < 100) {
	$server_port = 3600 + $server_port + 1;
	}
    $Cyc::debug = $debug;
    $cyc = Cyc->new (host => $host, port => $server_port);
    if ($cyc) {
	$OK = &TRUE;
	&cyc_function("(set-the-cyclist \"$cyclist\")") unless ($cyclist eq "");
    }
    ## &cyc_function("(in-package \"CYC\")");
    return ($OK);
}

# disconnect_cyc(): terminate the Cyc session
#
sub end_cyc {
    &debug_out(&TL_VERBOSE, "end_cyc(@_)\n");

    # Show the results of the profiling
    if ($profile) {
	printf STDERR "funct\t#calls\ttime\taverage\n";
	foreach my $key (sort(keys(%function_calls))) {
	    my($average) = ($function_calls{$key} > 0) ? ($function_times{$key} / $function_calls{$key}) : 0.0;
	    printf STDERR "%s\t%s\t%s\t%.3f\n", $key, $function_calls{$key}, $function_times{$key}, $average;
	}
    }
}

#........................................................................
# TODO: place cyc-related functions in Cyc.pm
# currently, code is copied from check_coverage.perl

# lispify(binding_array)
#
# Convert an array giving a binding value into a lisp list representation (String).
# In the case of NAT's or formulas, the value might have embedded array references
# which need to be recursively resolved.
#
# ex: ('#$fromLocation', ':ACTION', ':OBLIQUE-OBJECT) 
#     => "(#$fromLocation :ACTION :OBLIQUE-OBJECT)"
#
# ex: ('#$and', ARRAY(0x81c6ee4), ARRAY(0x81c6f98))
#         where ARRAY(0x81c6ee4) = ('#$isa', ':ACTION', '#$MovementProcess')
#           and ARRAY(0x81c6f98) = ('#$objectMoving', ':ACTION', ':SUBJECT')
#     => "(#$and (#$isa :ACTION #$MovementProcess) (#$objectMoving :ACTION :SUBJECT))"
#
sub lispify {
    my(@binding) = @_;
    my($value);
    my($result) = "(";
    &debug_out(&TL_VERY_VERBOSE, "lispify(@_)\n");

    foreach $value (@binding) {
	if ($value =~ /HASH\(0x/) {
	    # Ignore hash indicators 
	    # ex: Cyc=HASH(0x8194e8c) in (Cyc=HASH(0x8194e8c) #$hasMembers ?ORG :NOUN)
	    next;
	}
	elsif ($value =~ /ARRAY\(0x/) {
	    $result .= &lispify(@$value);
	}
	else {
	    $result .= $value;
	}
	$result .= " ";
    }
    $result =~ s/ $//;		# drop last space
    $result .= ")";

    return ($result);
}

# perlify_lisp_list(text): return Perl array with textual representation for each
# item in the input list
# NOTE: Embedded lists show up embedded in the outermost item
#
sub perlify_lisp_list {
    my($text) = @_;
    my(@result);
    &debug_out(&TL_VERBOSE, "perlify_lisp_list(@_)\n");

    if ($text =~ /\(/) {
	@result = @{scalar $cyc->lisp_read($text)};
	foreach my $item (@result) {
	    # Modify the element in place to show full representation
	    $item = &lispify(@$item) if ($item =~ /ARRAY\(0x/);
	}
    }
    &trace_array(\@result, &TL_VERBOSE, "perlify_lisp_list result");

    return (@result);
}

# escape_string(string): Returns string with lisp escape characters as needed
#
sub escape_string {
    my($string) = @_;

    # Make sure quote is escaped
    $string =~ s/\"/\\\"/g;

    # Make sure that backslashed are appropriately escaped
    $string =~ s/\\([^\"\\])/\\\\$1/g;	# backlash excluding \" and \\
    $string =~ s/([^\\])\\$/$1\\\\/;	# trailing backslash
    $string =~ s/^\\$/\\\\/;		# special case: single backslash

    return ($string);
}

# cyc_function(function_invocation)
#
# Use the TCP/IP port to send the function call to Cyc.
#
sub cyc_function {
    my($command) = @_;
    &debug_out(&TL_VERY_DETAILED, "cyc_function(%s)\n", join(",", @_));
    my($result) = "";
    my($start_time, $function);

    # Make note of the starting time if profiling
    # TODO: find out mor efficient way of determining the time
    if ($profile) {
	$start = &run_command("$gettimeofday");
    }

    # TODO: Make sure embedded strings are properly quoted
    # NOTE: This can't be reliably done after the fact
    # $command =~ s/([^\\])\"/$1\\\"/g;

    if (&trim($command) ne "") {
	$result = $cyc->fi_send($command);
	if ($result =~ /^ERROR 500/) {
	    &debug_out(&TL_DETAILED, "WARNING %s => %s", $command, $result);
	    $result = "NIL";
	}
	if ($uncyclify_results) {

	    $result = &uncyclify($result);
	}
    }
    else {
	&debug_out(&TL_BASIC, "WARNING: empty Cyc function call\n");
    }
    &debug_out(&TL_VERBOSE, "Cyc: %s => %s\n", $command, $result);

    # If profiling updating the counts associated with the calls
    if ($profile) {
	($function) = ($command =~ /^\s*\(\s*(\S+)/);
	my($end) = &run_command("$gettimeofday");
	&incr_entry(\%function_calls, $function);
	&incr_entry(\%function_times, $function, $end - $start);
    }

    return ($result);
}

sub cyc_function_fmt {
    my($command_fmt, @rest) = @_;
    &debug_out(&TL_VERY_DETAILED, "cyc_function_fmt(%s)\n", join(",", @_));
    return (&cyc_function(sprintf $command_fmt, @rest));
}

# cyc_eval(expression): Send the expression to the Cyc server for 
# evaluation and return the result.
#
sub cyc_eval {
    return (&cyc_function("(progn @_)"));
}

# cyc_list_function(function_invocation)
#
# Send a function invocation to Cyc and convert the result to Perl list.
# NOTE: Embedded lists show up using array references.
#
sub cyc_list_function {
    my($command) = @_;
    &debug_out(&TL_VERBOSE, "cyc_list_function(%s)\n", join(",", @_));
    
    my($result) = &cyc_function($command);
    $result =~ s/(#<)|(>)//g;
    my(@list) = @{scalar $cyc->lisp_read($result)};
    &debug_out(&TL_VERY_VERBOSE, "cyc_list_function => (%s)\n", join(", ", @list));

    return (@list);
}

sub cyc_list_function_fmt {
    my($command_fmt, @rest) = @_;
    return (&cyc_list_function(sprintf $command_fmt, @rest));
}

sub cyc_string_list_function {
    my($command) = @_;
    &debug_out(&TL_VERBOSE, "cyc_string_list_function(%s)\n", join(",", @_));

    my(@list) = &cyc_list_function($command);
    foreach my $item (@list) {
	$item =~ s/\\\"/\"/g;
	$item =~ s/^\"(.*)\"$/$1/;
    }
    &debug_out(&TL_VERY_VERBOSE, "cyc_string_list_function => (%s)\n", join(", ", @list));

    return (@list);
}

sub cyc_string_list_function_fmt {
    my($command_fmt, @rest) = @_;
    return (&cyc_string_list_function(sprintf $command_fmt, @rest));
}

# cyc_string_function(function_invocation)
#
# Send a function invocation to Cyc and convert the result to Perl string.
#
sub cyc_string_function {
    my($command) = @_;
    &debug_out(&TL_VERBOSE, "cyc_string_function(%s)\n", join(",", @_));
    
    my($result) = &cyc_function($command);
    $result = "" if ($result eq "NIL");
    $result =~ s/\\\"/\"/g;
    $result =~ s/^\"(.*)\"$/$1/;

    return ($result);
}

sub cyc_string_function_fmt {
    my($command_fmt, @rest) = @_;
    return (&cyc_string_function(sprintf $command_fmt, @rest));
}

# cyc_meaning_function(function_invocation)
#
# Same as the above but embedded lists show up using Lisp notation.
#
sub cyc_meaning_function {
    my($command) = @_;
    &debug_out(&TL_VERY_DETAILED, "cyc_meaning_function(%s)\n", join(",", @_));

    my(@list) = &cyc_list_function($command);
    &trace_array(\@list);
    my($meaning);
    foreach $meaning (@list) {
	# Modify the element in place to show full representation
	$meaning = &lispify(@$meaning) if ($meaning =~ /ARRAY\(0x/);
    }

    return (@list);
}

sub cyc_meaning_function_fmt {
    my($command_fmt, @rest) = @_;
    &debug_out(&TL_VERY_DETAILED, "cyc_meaning_function_fmt(%s)\n", join(",", @_));
    return (&cyc_meaning_function(sprintf $command_fmt, @rest));
}

# cyc_boolean_function(command): Invoke the Cyc function returning a boolean value.
#
sub cyc_boolean_function {
    return (&cyc_function(@_) ne "NIL");
}

sub cyc_boolean_function_fmt {
    my($command_fmt, @rest) = @_;
    return (&cyc_boolean_function(sprintf $command_fmt, @rest));
}

# uncyclify(expression)
#
# Remove the #$ prefixes, etc. from the Cyc expression
# ex: "#$Monkey" => "Monkey"; "#<(#$JuvenileFn #$Person)>" => "(JuvenileFn Person)"
# TODO: eliminate the loop, applying the pattern globally
#
sub uncyclify {
    my($expression) = @_;

    $expression =~ s/\#\$//g;
    while ($expression =~ /\#\<([^\<\>]*)\>/) {
	$expression =~ s/\#\<([^\<\>]*)\>/$1/;
    }
    &debug_out(&TL_VERY_DETAILED, "uncyclify(%s) => %s\n", join(" ", @_), $expression);

    return ($expression);
}

# isa($term, $type) => boolean
#
# Returns true if (isa Term Type) holds in any MT in Cyc
#
sub isa {
    &debug_out(&TL_VERBOSE, "isa(@_)\n");
    local ($term, $type) = @_;
    $term =~ s/\"/\\"/g;

    my($function) = sprintf "(isa-in-any-mt? (fort-for-string \"%s\") (fort-for-string \"%s\"))", $term, $type;
    return (&cyc_function($function) eq "T");
}

# genls($term, $type) => boolean
#
# Returns true if (genls Term Type) holds in any MT in Cyc
# TODO: genls-in-any-mt?
#
sub genls {
    &debug_out(&TL_VERBOSE, "genls(@_)\n");
    local ($term, $type) = @_;
    $term =~ s/\"/\\"/g;

    my($function) = sprintf "(with-all-mts (genls? (fort-for-string \"%s\") (fort-for-string \"%s\")))", $term, $type;
    return (&cyc_function($function) eq "T");
}

# generality_estimate(term): Returns Cyc's generality estimate for the term
#
sub generality_estimate {
    &debug_out(&TL_VERBOSE, "generality_estimate(@_)\n");
    my($term) = @_;
    $term =~ s/\"/\\"/g;

    my($estimate) = &cyc_function_fmt("(generality-estimate (fort-for-string \"%s\"))",
				      $term);
    $estimate = 0 if ($estimate eq "NIL");
    return ($estimate);
}

# load_supporting_code(support_code_function, [support_code_filename])
#
# Load in the supporting lisp code unless the supporting function is already defined.
# The name for the lisp file defaults to the name of the script with the extension
# changes from .perl to .lisp.
# TODO: just change whatever extension is present to .lisp
#
sub load_supporting_code {
    my($support_function, $support_code) = @_;
    my($OK) = &TRUE;
    if (!defined($support_code)) {
	$support_code = &basename($0, ".perl") . ".lisp";
    }
    $support_code = &make_full_path($support_code);

    &reference_var(*reload);
    if ($reload ||
	&cyc_function("(ignore-errors (type-of \#\'$support_function))") !~ /FUNCTION/) {

	# Make sure Allegro can process the Cyc constants
	## &cyc_function("\#+Allegro (cl:set-dispatch-macro-character \#\\\# \#\\\$ \'sli::read-constant cl:*readtable*) (+ 2 2)");
	## &cyc_function("(pwhen (member :ALLEGRO *features*) (cl:set-dispatch-macro-character \#\\\# \#\\\$ \'sli::read-constant cl:*readtable*))");
	&cyc_function("(progn \#+Allegro (cl:set-dispatch-macro-character \#\\\# \#\\\$ \'sli::read-constant cl:*readtable*))");

        # downcase the directory to circumvent stupid Cyc Allegro bug
        # NOTE: there must be a symbolic link with the downcase version
        ## $support_code =~ s/(.*)\/([^\/]+).lisp/\L$1\E\/$2.lisp/;

	# Load in the file
	my($result) = &cyc_function("(load-api-internal \"$support_code\")");
	$OK = ($result eq "T");

	# Perform sanity check that the function actually exists
	if (&cyc_function("(ignore-errors (type-of \#\'$support_function))") !~ /FUNCTION/) {
	    &error_out("Unable to load in supporting code: $support_code\n");
	}
    }

    return ($OK);
}

# valid_fort(term): Determines whether the term represents a valid fort
#
sub valid_fort {
    my($term) = @_;
    &debug_out(&TL_VERBOSE, "valid_fort(%s)\n", join(" ", @_));

    $term = &escape_string($term);
    my($command) = sprintf "(fort-p (fort-for-string \"%s\"))", $term;
    my($result) = &cyc_function($command);
    my($OK) = ($result eq "T");

    return ($OK);
}

# set_the_cyclist([new-cyclist=cyclist]): Set cyclist to ID for current user
# NOTE: uses global $cyclist and $user variables
#
sub set_the_cyclist {
    &debug_out(&TL_VERBOSE, "set_the_cyclist(@_)\n");
    my($new_cyclist);
    $new_cyclist = $cyclist if (! defined($new_cyclist));

    if ($new_cyclist eq "") {
        $new_cyclist = &cyc_function("(first (return-list-of-constants '(\#\$eMailAddressText ?W \"$user\@cyc.com\") \#\$InferencePSC))");
	if ($new_cyclist eq "NIL") {
	    print STDERR "Warning: no eMailAddressText assertion for $user\n";
	}
    }
    if ($new_cyclist ne "") {
        &cyc_function_fmt("(set-the-cyclist (fort-for-string \"%s\"))", $new_cyclist);
    }
}

# function_over_term(function, term): format a lisp call to the given function using
# the Cyc term as an argument.
# NOTE: This is used in the various Cyc function wrappers (e.g., cyc_list_function)
# ex: 'function_over_term("all-isa", "Dog")' => '(all-isa (gb-guess-fort "Dog"))'
#
sub function_over_term {
    my($function, $term) = @_;
    $term = &escape_string($term);
    return "($function (fort-for-string \"$term\"))";
}

# main_fort_type(term): returns the main type of the Cyc term
# TODO: use the most general type in cases with more than one
# ex: main_fort_type(Query) => AttributeValue
# NOTE: this version doesn't (yet) work in OpenCyc
#
sub old_main_fort_type {
    my($term) = @_;
    $term = &escape_string($term);
    return (&cyc_function_fmt("(first (with-all-mts (fort-types (fort-for-string \"%s\"))))", 
			      $term));
}

# main_fort_type(term): returns the type of the Cyc term using one of the main categories:
#     Function-Denotational, Predicate, Collection, AttributeValue
# ex: fort_type(Query) => AttributeValue
#
my(%main_fort_types);
#
sub main_fort_type {
    my($term) = @_;
    my($fort_type) = "Individual";

    # Initialize the hash table with the main types
    if (!defined($main_fort_types{"Collection"})) {
	map { $main_fort_types{$_} = &TRUE;  }  ("Function-Denotational", "Collection", "Predicate", "AttributeValue");
    }

    # See if any of the term types are in the main types
    my(@term_types) = &term_types($term);
    foreach my $type (@term_types) {
	if (defined($main_fort_types{$type})) {
	    $fort_type = $type;
	    last;
	}
    }
    &debug_out(&TL_VERBOSE, "main_fort_type(@_) => $fort_type\n");

    return ($fort_type);
}

# eval_cyc_meaning_function: Evaluate denotation-type function applied to the term
# (with proper lisp escaping applied, such as for embedded quotes).
#
sub eval_cyc_meaning_function {
    my($function, $term) = @_;
    return (&cyc_meaning_function(&function_over_term($function, $term)));
}

# term_types(term): returns all cyc types associated with term
#
sub term_types {
    return (&eval_cyc_meaning_function("all-isa-in-any-mt", @_));
}

# direct_term_types(term): return types explicitly asserted for term (i.e.,
# isa's w/o transitivity)
#
sub direct_term_types {
    my($term) = @_;
    $term = &escape_string($term);
    return (&cyc_meaning_function_fmt("(pred-values-in-any-mt (fort-for-string \"%s\") #\$isa)", 
				      $term));
}

# generalization_pred(term): returns CycL predicate for generalization of the term based
# on its type, with the default being genls
# ex: generalization_pred(Mammal) => genls
# ex: generalization_pred(Query) => genlAttributes
#
my(%generalization_pred);
#
sub generalization_pred {
    my($term) = @_;

    if (!defined($generalization_pred{"Collection"})) {
	%generalization_pred = ( "Collection" => "genls",
				 "Function-Denotational" => "",		# ???
				 "Predicate" => "genlPreds",
				 "AttributeValue" => "genlAttributes",
				 );
    }

    my($type) = &main_fort_type($term);
    my($pred) = $generalization_pred{$type} || "genls";

    return ($pred);
}

# all_generalizations_function(term): returns SubL function for obtaining all generalizations of term based
# on its type, with the default being all-genls-in-any-mt
# ex: all_generalizations_function(Mammal) => all-genls-in-any-mt
# ex: all_generalizations_function(Query) => all-genlAttributes-in-any-mt
# TODO: abstract this and generalization_pred into a more general function
#
my(%all_generalizations_function);
#
sub all_generalizations_function {
    my($term) = @_;

    if (!defined($all_generalizations_function{"Collection"})) {
	%all_generalizations_function = ( "Collection" => "all-genls-in-any-mt",
					  "Function-Denotational" => "",		# ???
					  "Predicate" => "all-genl-preds",
					  "AttributeValue" => "all-genl-attributes-in-any-mt",
					  );
    }

    my($type) = &main_fort_type($term);
    my($function) = $all_generalizations_function{$type} || "all-genls-in-any-mt";

    return ($function);
}

# term_generalizations(term): returns all cyc collections that generalize the cyc term
# note: the term itself is not returned in the list
# TODO: define an efficient function to remove an element from a list
#
sub term_generalizations {
    my($term) = @_;
    my($function) = &all_generalizations_function($term);
    my(@generalizations) = &eval_cyc_meaning_function($function, @_);
    my($discarded_term) = pop(@generalizations);
    &assert($discarded_term eq $term) unless (!defined($discarded_term));

    return (@generalizations);
}

# direct_term_generalizations(term): return types explicitly asserted for term (i.e.,
# genls's w/o transitivity)
# ex: direct_term_generalizations(Dog) => (Animal CanineAnimal DomesticatedAnimal)
#
sub direct_term_generalizations {
    my($term) = @_;
    my($pred) = &generalization_pred($term);
    $term = &escape_string($term);
    return (&cyc_meaning_function_fmt("(pred-values-in-any-mt (fort-for-string \"%s\") #\$%s)", 
				      $term, $pred));
}


# ------------------------------------------------------------------------

# Indicate that this loaded OK
1;
