# *-*-perl-*-*
eval 'exec perl -Ssw $0 "$@"'
    if 0;
#!/usr/bin/perl -sw
#
# feature2ml.perl: convert from tab-delimited feature specification to format
# used in the ML repository  (c4.5-based). Alternatively, the format used
# by Weka is also supported.
#
# TODO:
# - use process_input_files (see extract_cooccuring_forts.perl)
# - add brief documentation on the support formats (e.g., c45 & weka)
#

# Load in the common module, making sure the script dir is in the Perl lib path
BEGIN { my $dir = `dirname $0`; chomp $dir; unshift(@INC, $dir); }
require 'common.perl';

# Parse the command-line arguments
if (!defined($ARGV[0])) {
    my($options) = "options = [-nolabels] [-class_pos] [-first] [-dataset=name] [-preserve]  [-arff]";
    $note = "NOTE:\n\nFor c4.5, creates <dataset_name>.names and <dataset_name>.data\n";
    $note .= "For Weka, creates .arff file/\n";

    my($example) = "examples:\n\n$script_name compoundString < compoundString.table\n\n";
    $example .= "$0 -d=5 test_english.table\n\n";
    $example .= "$0 multiword.position.table\n\n";
    $example .= "$0 -first -weka ai-restaurant-example.features\n";
    $example .= "less ai-restaurant-example.arff\n\n";
    $example .= "foreach.perl -trace '$0 \$f' *.features >| feature2ml.log 2>&1\n\n";

    die "\nusage: $script_name [options] dataset_file\n\n$options\n\n$note\n\n$example\n";
}

# Check command-line options
&init_var(*arff, &FALSE);	# alias for -weka
&init_var(*weka, $arff);	# use Weka's .arff format
&init_var(*c45, ! $weka);	# use C4.5 .names/.data format
&init_var(*dataset, "");	# name of dataset
&init_var(*nolabels, &FALSE);	# no labels in data file
&init_var(*labels, "");		# labels for the features
&init_var(*first, &FALSE);	# class value given first
&init_var(*class_pos, ($first ? 0 : -1));	# 0-based position of the classificaton variable

# Derive the dataset name from the file if not given
if ($dataset eq "") {
    $dataset = $ARGV[0];
    if ($dataset =~ /(.*)\.(table|features?)$/) {
	# Use the basename of the table file as the dataset name
	$dataset = $1;
    }
}

local(@labels);
local(@value_hashes);
local($num_features) = 0;
local(@feature_data);

# Extract the feature values for each training instance, making note of the
# values that occur for each feature variable.
#
my($count) = 0;
READ_DATA:
while (<>) {
    &dump_line();
    chomp;
    s/\r//;
    next if (/\s*;/);		# ignore comments

    # Extract features
    my(@features) = split(/\t/, $_);
    $count++;

    # Assign the labels if first line
    if ($count == 1) {
	# Usually, the first row specifies the feature labels
	if ($nolabels == &FALSE) {
	    &init_features(@features);
	    next;
	}
	# If not, use the labels specified on command-line (or a generic "Fn" label)
	else {
	    my(@temp_labels) = &tokenize($labels);
	    for ($i = 0; $i <= $#features; $i++) {
		$temp_labels[$i] = sprintf "F%d", $i unless defined($temp_labels[$i]);
	    }
	    &init_features(@temp_labels);
	}
    }

    # Make note of the feature values used
    # TODO: account for embedded commas, etc.
    my($i);
    for ($i = 0; $i < $num_features; $i++) {
	if (!defined($features[$i])) {
	    &error_out("Incomplete feature set: $_\n");
	    next READ_DATA;
	}
	$features[$i] = "\'" . $features[$i] . "\'" if ($features[$i] !~ /^\w+$/);
	&add_feature_value($i, $features[$i]);
    }

    # Reorganize the feature values, so that the class value occurs last
    if ($class_pos != -1) {
	my($class_value) = $features[$class_pos];
	splice(@features, $class_pos, 1);
	push(@features, $class_value);
    }

    # Add the feature values to the accumulated data 
    # TODO: just store the array of features instead of a string list
    my($data_spec) = join(", ", @features);
    push(@feature_data, $data_spec);
}

# Prepare output data file(s)
#
if ($weka) {
    &output_arff_description();}
elsif ($c45) {
    &output_c45_description();
}
else {
    &error("ERROR: neither C45 nor Weka specified\n");
}


&exit();

#------------------------------------------------------------------------

# init_features(label_array_ref)
#
# Initialize the anonymous associative arrays in @values_hashes used
# for storing the values encountered for each feature.
#
sub init_features {
    my(@features) = @_;
    debug_out(5, "init_features(@_)\n");

    @labels = @features;
    $num_features = (1 + $#labels);
    ## @value_hashes = ({ }) x $num_features;
    my($i);
    for ($i = 0; $i < $num_features; $i++) {
	$value_hashes[$i] = { };
    }
    &trace_array(\@value_hashes);
    if ($class_pos < 0) {
	$class_pos = $#labels;
    }
}

# add_feature_value(feature_pos, value)
#
# Add the value to the table of values for the feature. This uses the
# anonymous associative arrays stored in @value_hashes.
#
sub add_feature_value {
    my($pos, $value) = @_;
    &debug_out(6, "add_feature_value(@_)\n");

    local(*hash) = $value_hashes[$pos];
    &incr_entry(*hash, $value);
}

# get_feature_values(feature_pos)
#
# Return a list of the values that occured for the feature. Again,
# this uses the anonymous associative arrays stored in @value_hashes.
#
sub get_feature_values {
    my($pos) = @_;
    &debug_out(6, "get_feature_values(@_)\n");
    my(@values);

    my($hash_ref) = $value_hashes[$pos];
    if (defined($hash_ref)) {
	@values = keys(%$hash_ref);
    }

    return (@values);
}

# output_c45_description(): produces .names and .data file for C4.5
# TODO: add support for creating a .test file as well
# TODO: parameterize to eliminate use of globals
#
sub output_c45_description {
    # Prepare the feature description file (<dataset>.names)
    #
    my($description) = "";
    $description .= sprintf "%s.\n", join(",", &get_feature_values($class_pos));
    for ($i = 0; $i < $num_features; $i++) {
	if ($i != $class_pos) {
	    $description .= sprintf "%s: %s.\n", 
	    $labels[$i], join(", ", &get_feature_values($i));
	}
    }
    &write_file("$dataset.names", $description);

    # Output the data file (<dataset>.data)
    #
    ## my($data) = join("\n", @feature_data) . "\n";
    ## &write_file("$dataset.data", $data);
    # TODO: prepare a variant of write_file that takes a reference to the
    # lines for efficiency
    #
    open(DATA, ">$dataset.data") or
	die ("Unable to create data file $dataset.data ($!)\n");
    map { print DATA "$_.\n"; } @feature_data;
    close(DATA);

    print "Output created:\n";
    print "$dataset.names describes the features\n";
    print "$dataset.data contains the data\n";
}


# output_arff_description(): produces .arff file for Weka
# TODO: parameterize to eliminate use of globals
#
sub output_arff_description {
    # Output the data file (<dataset>.arff)
    #
    open(ARFF, ">$dataset.arff") or
	die ("Unable to create Weka data file $dataset.arff ($!)\n");
    print ARFF "\@relation $dataset\n";

    # Output the attribute specification
    my($i);
    for ($i = 0; $i < $num_features; $i++) {
	if ($i != $class_pos) {
	    printf ARFF "\@attribute %s { %s }\n", $labels[$i], join(", ", &get_feature_values($i));
	}
    }
    printf ARFF "\@attribute class { %s }\n", join(", ", &get_feature_values($class_pos));

    # Output the data points
    print ARFF "\@data\n";
    map { print ARFF "$_\n"; } @feature_data;

    close(ARFF);

    print "Output created:\n";
    print "$dataset.arff contains the data and also the feature specification\n";
}
